/*
 * Copyright  Micromega Corporation 2002-2007. All rights reserved.
 */
package com.micromegacorp.math.v3_spi;
/**
 * @file    Float32.java
 * @target  Javelin Stamp
 *
 * This class provides support for 32-bit floating point numbers which conform
 * to IEEE standard 754 (the 32-bit floating format used by standard Java).
 * The uMFPU V3 floating point coprocessor is used to perform the math, since
 * the Javelin Stamp does not have built-in support for floating point.
 *
 * Floating point values are stored in two 16-bit integers in the
 * following format (the exponent has a bias of 127):
 * <pre>
 *               high                              low
 * |=|===============|=============| |===============================|
 * |S|<---Exponent-->|<------------| |-----Mantissa----------------->|
 * |=|===============|=============| |===============================|
 *  1          8            7                        16              #bits
 *
 * e.g.   1.0 = 0x3F80, 0x0000
 *       -1.0 = 0xBF80, 0x0000
 *       10.0 = 0x4120, 0x0000
 *       0.01 = 0x3C23, 0xD70A
 * </pre>Additional information regarding floating point numbers can be
 * obtained by reviewing the IEEE 754 standard or standard Java documentation.

 * @author Cam Thompson, Micromega Corporation
 * @version
 *    November 2, 2006
 *    - high, low values of Float32 made public
 *    March 20, 2006
 *    - original version for uM-FPU V3
 */

public class Float32 {
  /*
   * high 16 bits of 32-bit value.
   */
  public int high;
  /*
   * low 16 bits of 32-bit value.
   */
  public int low;

  // special constants (NaN, +infinity, -infinity)
  private final static int NAN_HIGH   = (short)0x7FFF;
  private final static int NAN_LOW    = (short)0xFFFF;
  protected final static String notReset = "?FPU not reset";
  protected static StringBuffer sbuf = new StringBuffer(10);

  //-------------------- constructors -----------------------------------------

  /**
   * Create a new FpuData object initialized to zero.
   */
  public Float32() {
    high = 0;
    low = 0;
  }
  /**
   * Create a new <code>Float32</code> initialized to the value of another
   * <code>Float32</code>.
   * @param fnum <code>Float32</code> number.
   */
  public Float32(Float32 fnum) {
    set(fnum);
  }
  /**
   * Create a new <code>Float32</code> initialized to the floating point value
   * contained in two 16-bit integers.
   * @param high high 16 bits of floating point value.
   * @param low low 16 bits of floating point value.
   */
  public Float32 (int high, int low) {
    set(high, low);
  }
  /**
   * Create a new <code>Float32</code> initialized to the value of a 16-bit
   * integer.
   * @param num 16-bit integer value (sign extended).
   */
  public Float32(int num) {
    set(num);
  }

  //-------------------- set --------------------------------------------------
  /**
   * Set <code>Float32</code> to the value of another <code>Float32</code>.
   * @param fnum <code>Float32</code> number.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 set(Float32 fnum) {
    this.high = fnum.high;
    this.low = fnum.low;
    return(this);
  }
  /**
   * Set <code>Float32</code> to the floating point value contained in two
   * 16-bit integers.
   * @param high high 16 bits of floating point value.
   * @param low low 16 bits of floating point value.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 set(int high, int low) {
    this.high = high;
    this.low = low;
    return(this);
  }
  /**
   * Set <code>Float32</code> to the value of a 16-bit integer.
   * @param num 16-bit integer value (sign extended).
   * @return <code>Float32</code> reference to result.
   */
  public Float32 set(int num) {
    if (num == 0) {
      high = 0;
      low = 0;
    }
    else if (Fpu.fpuClockPin == 0) {
      // FPU has not been reset
      high = NAN_HIGH;
      low = NAN_LOW;
    }
    else {
      Fpu.startWrite();
      Fpu.write(Fpu.LOADWORD);
      Fpu.writeWord(num);
      Fpu.wait();
      Fpu.startWrite();
      Fpu.write(Fpu.FREAD0);
      high = Fpu.readWord();
      low = Fpu.readWord();
    }
    return(this);
  }
  /**
   * Set <code>Float32</code> to the floating point value represented by a
   * <code>String</code>.
   * @param s <code>String</code> representing the floating point value.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 set(String s) {
    if (Fpu.fpuClockPin == 0) {
      // FPU has not been reset
      high = NAN_HIGH;
      low = NAN_LOW;
    }
    else {
      Fpu.startWrite();
      Fpu.write(Fpu.ATOF);
      Fpu.writeString(s);
      Fpu.wait();
      Fpu.startWrite();
      Fpu.write(Fpu.FREAD0);
      high = Fpu.readWord();
      low = Fpu.readWord();
    }
    return(this);
  }
  /**
   * Set <code>Float32</code> to the floating point value represented by a
   * <code>StringBuffer</code>.
   * @param sb <code>StringBuffer</code> representing the floating point value.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 set(StringBuffer sb) {
    return(set(sb.toString()));
  } // end set

  //-------------------- add --------------------------------------------------
  /**
   * Add another <code>Float32</code> to this <code>Float32</code>.
   * @param fnum <code>Float32</code> to add.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 add(Float32 fnum) {
    return(add(fnum.high, fnum.low));
  }
  /**
   * Add the floating point value contained in two 16-bit integers to this
   * <code>Float32</code>.
   * @param high high 16 bits of floating point value to add.
   * @param low low 16 bits of floating point value to add.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 add(int high, int low) {
    loadA();
    loadB(high, low);
    return(saveA(Fpu.FADD0));
  }
  /**
   * Add a 16-bit integer value to this <code>Float32</code>.
   * @param num 16-bit integer value (sign extended) to add.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 add(int num) {
    loadA();
    loadC(num);
    return(saveA(Fpu.FADD0));
  }

  //-------------------- subtract ---------------------------------------------
  /**
   * Subtract another <code>Float32</code> from this <code>Float32</code>.
   * @param fnum <code>Float32</code> to subtract.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 subtract(Float32 fnum) {
    return(subtract(fnum.high, fnum.low));
  }
  /**
   * Subtract the floating point value contained in two 16-bit integers from
   * this <code>Float32</code>.
   * @param high high 16 bits of floating point value to subtract.
   * @param low low 16 bits of floating point value to subtract.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 subtract(int high, int low) {
    loadA();
    loadB(high, low);
    return(saveA(Fpu.FSUB0));
  }
  /**
   * Subtract a 16-bit integer from this <code>Float32</code>.
   * @param num 16-bit integer value (sign extended) to subtract.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 subtract(int num) {
    loadA();
    loadC(num);
    return(saveA(Fpu.FSUB0));
  }

  //-------------------- multiply ---------------------------------------------
  /**
   * Multiply this <code>Float32</code> by another <code>Float32</code>.
   * @param fnum <code>Float32</code> multiplier.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 multiply(Float32 fnum) {
    return(multiply(fnum.high, fnum.low));
  }
  /**
   * Multiply this <code>Float32</code> by the floating point value contained
   * in two 16-bit integers.
   * @param high high 16 bits of floating point multiplier.
   * @param low low 16 bits of floating point multiplier.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 multiply(int high, int low) {
    loadA();
    loadB(high, low);
    return(saveA(Fpu.FMUL0));
  }
  /**
   * Multiply this <code>Float32</code> by a 16-bit integer.
   * @param num 16-bit integer (sign extended) multiplier.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 multiply(int num) {
    loadA();
    loadC(num);
    return(saveA(Fpu.FMUL0));
  }

  //-------------------- divide -----------------------------------------------
  /**
   * Divide this <code>Float32</code> by another <code>Float32</code>.
   * @param fnum <code>Float32</code> divisor.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 divide(Float32 fnum) {
    return(divide(fnum.high, fnum.low));
  }
 /**
   * Divide this <code>Float32</code> by the floating point value contained
   * in two 16-bit integers.
   * @param high high 16 bits of floating point divisor.
   * @param low low 16 bits of floating point divisor.
   * @return <code>Float32</code> reference to result.
   */
  public Float32 divide(int high, int low) {
    loadA();
    loadB(high, low);
    return(saveA(Fpu.FDIV0));
  }
  /**
   * Divide this floating point number by a 16-bit integer.
   * @param num 16-bit integer divisor (sign extended).
   * @return <code>Float32</code> reference to result.
   */
  public Float32 divide(int num) {
    loadA();
    loadC(num);
    return(saveA(Fpu.FDIV0));
  }

  //-------------------- toString ---------------------------------------------
  /**
   * Convert value of this <code>Float32</code> to a <code>String</code>.
   * @return <code>String</code> representation of <code>Float32</code>.
   */
  public String toString() {
    if (Fpu.fpuClockPin == 0)
      return(notReset);
    else {
      Fpu.startWrite();
      Fpu.write(Fpu.LEFT, Fpu.FWRITEA);
      Fpu.writeWord(high);
      Fpu.writeWord(low);
      Fpu.write(Fpu.FTOA, 0, Fpu.RIGHT);
      return(Fpu.readString());
    }
  }

  //-------------------- toHexString ------------------------------------------
  /**
   * Convert <code>FpuData</code> value to a hexadecimal string.
   * @return text representation of  <code>FpuData</code> value.
  */
  public String toHexString() {
    // create hex string
    sbuf.clear();
    sbuf.append("0x");
    sbuf.append(toHex(high >> 12));
    sbuf.append(toHex(high >> 8));
    sbuf.append(toHex(high >> 4));
    sbuf.append(toHex(high));
    sbuf.append(toHex(low >> 12));
    sbuf.append(toHex(low >> 8));
    sbuf.append(toHex(low >> 4));
    sbuf.append(toHex(low));
    return (sbuf.toString());
  }

  //-------------------- toHex ------------------------------------------------

  protected void loadA() {
    Fpu.startWrite();
    Fpu.write(Fpu.LEFT, Fpu.FWRITEA);
    Fpu.writeWord(this.high);
    Fpu.writeWord(this.low);
  }

  protected void loadB(int high, int low) {
    Fpu.write(Fpu.FWRITE0);
    Fpu.writeWord(high);
    Fpu.writeWord(low);
  }

  protected void loadC(int num) {
    Fpu.write(Fpu.LOADWORD);
    Fpu.writeWord(num);
  }

  protected Float32 saveA(int opcode) {
    Fpu.write(opcode, Fpu.RIGHT);
    Fpu.wait();
    Fpu.write(Fpu.FREAD0);
    this.high = Fpu.readWord();
    this.low = Fpu.readWord();
    return(this);
  }

  protected char toHex(int n) {
    char hexDigit;
    hexDigit = (char) ('0' + (n & 0x0F));
    if (hexDigit > '9') hexDigit += 7;
    return(hexDigit);
  }
}