/*
 * Copyright  Micromega Corporation 2002-2007. All rights reserved.
 */
package com.micromegacorp.math.v3_spi;
/**
 * @file    Int32.java
 * @target  Javelin Stamp
 *
 * This class provides support for 32-bit floating point numbers which conform
 * to IEEE standard 754 (the 32-bit floating format used by standard Java).
 * The uMFPU V3 floating point coprocessor is used to perform the math, since
 * the Javelin Stamp does not have built-in support for floating point.
 *
 * Floating point values are stored in two 16-bit integers in the
 * following format (the exponent has a bias of 127):
 * <pre>
 *               high                              low
 * |=|===============|=============| |===============================|
 * |S|<---Exponent-->|<------------| |-----Mantissa----------------->|
 * |=|===============|=============| |===============================|
 *  1          8            7                        16              #bits
 *
 * e.g.   1.0 = 0x3F80, 0x0000
 *       -1.0 = 0xBF80, 0x0000
 *       10.0 = 0x4120, 0x0000
 *       0.01 = 0x3C23, 0xD70A
 * </pre>Additional information regarding floating point numbers can be
 * obtained by reviewing the IEEE 754 standard or standard Java documentation.

 * @author Cam Thompson, Micromega Corporation
 * @version
 *    November 2, 2006
 *    - original version for uM-FPU V3
 */

public class Int32 {
  /*
   * high 16 bits of 32-bit value.
   */
  public int high;
  /*
   * low 16 bits of 32-bit value.
   */
  public int low;

  // special constants
  private final static int NAN_HIGH   = (short)0x7FFF;
  private final static int NAN_LOW    = (short)0xFFFF;
  protected final static String notReset = "?FPU not reset";
  protected static StringBuffer sbuf = new StringBuffer(10);

    //-------------------- constructors -----------------------------------------

  /**
   * Create a new FpuData object initialized to zero.
   */
  public Int32() {
    high = 0;
    low = 0;
  }
  /**
   * Create a new <code>Intt32</code> initialized to the value of another
   * <code>Int32</code>.
   * @param fnum <code>Int32</code> number.
   */
  public Int32(Int32 fnum) {
    set(fnum);
  }
  /**
   * Create a new <code>Int32</code> initialized to the floating point value
   * contained in two 16-bit integers.
   * @param high high 16 bits of floating point value.
   * @param low low 16 bits of floating point value.
   */
  public Int32 (int high, int low) {
    set(high, low);
  }
  /**
   * Create a new <code>Int32</code> initialized to the value of a 16-bit
   * integer.
   * @param num 16-bit integer value (sign extended).
   */
  public Int32(int num) {
    set(num);
  }

  //-------------------- set --------------------------------------------------
  /**
   * Set <code>Int32</code> to the value of another <code>Int32</code>.
   * @param fnum <code>Int32</code> number.
   * @return <code>Int32</code> reference to result.
   */
  public Int32 set(Int32 fnum) {
    this.high = fnum.high;
    this.low = fnum.low;
    return(this);
  }
  /**
   * Set <code>Int32</code> to the floating point value contained in two
   * 16-bit integers.
   * @param high high 16 bits of floating point value.
   * @param low low 16 bits of floating point value.
   * @return <code>Int32</code> reference to result.
   */
  public Int32 set(int high, int low) {
    this.high = high;
    this.low = low;
    return(this);
  }
  /**
   * Set <code>Int32</code> to the value of a 16-bit integer.
   * @param num 16-bit integer value (sign extended).
   * @return <code>Float32</code> reference to result.
   */
  public Int32 set(int num) {
    if (num == 0) {
      high = 0;
      low = 0;
    }
    else if (Fpu.fpuClockPin == 0) {
      // FPU has not been reset
      high = NAN_HIGH;
      low = NAN_LOW;
    }
    else {
      Fpu.startWrite();
      Fpu.write(Fpu.LONGWORD);
      Fpu.writeWord(num);
      Fpu.wait();
      Fpu.startWrite();
      Fpu.write(Fpu.LREAD0);
      high = Fpu.readWord();
      low = Fpu.readWord();
    }
    return(this);
  }
  /**
   * Set <code>int32</code> to the floating point value represented by a
   * <code>String</code>.
   * @param s <code>String</code> representing the floating point value.
   * @return <code>int32</code> reference to result.
   */
  public Int32 set(String s) {
    if (Fpu.fpuClockPin == 0) {
      // FPU has not been reset
      high = NAN_HIGH;
      low = NAN_LOW;
    }
    else {
      Fpu.startWrite();
      Fpu.write(Fpu.ATOL);
      Fpu.writeString(s);
      Fpu.wait();
      Fpu.startWrite();
      Fpu.write(Fpu.LREAD0);
      high = Fpu.readWord();
      low = Fpu.readWord();
    }
    return(this);
  }
  /**
   * Set <code>int32</code> to the floating point value represented by a
   * <code>StringBuffer</code>.
   * @param sb <code>StringBuffer</code> representing the floating point value.
   * @return <code>int32</code> reference to result.
   */
  public Int32 set(StringBuffer sb) {
    return(set(sb.toString()));
  } // end set
}